#include<string.h>
#include<stdlib.h>
#include"argon.h"

#include"GL_helper.h"
#include"HD_helper.h"
#include"calibration.h"

#include<unistd.h>
#include<time.h>



static HHD ghHD = HD_INVALID_HANDLE;

static HDSchedulerHandle hUpdateDeviceCallback;

static HDdouble workspacemodel[16];
static double kCursorScreenSize = 20.0;
static double gCursorScale = 1.0;

double maxForce=1;
double k=1;

float atomRadius=0.1;

int activeAtom=-1;

void displayFunction(void);

void updateWorkspace();


HapticDisplayState deviceState;

/*******************************************************************************
 This handler will get called when the application is exiting. This is our
 opportunity to cleanly shutdown the HD API.
*******************************************************************************/
void exitHandler(void)
{
    hdStopScheduler();
    hdUnschedule(hUpdateDeviceCallback);

	printf("exitting\n");

    if (ghHD != HD_INVALID_HANDLE)
    {
        hdDisableDevice(ghHD);
        ghHD = HD_INVALID_HANDLE;
    }

}

HDCallbackCode HDCALLBACK copyHapticDisplayState(void *pUserData)
/*******************************************************************************
 This handler copies deviceState to user space.
*******************************************************************************/
{
	memcpy(pUserData, &deviceState, sizeof(deviceState));

    return HD_CALLBACK_DONE;
}


HDCallbackCode HDCALLBACK setHapticUserForce(void *pUserData)
/*******************************************************************************
 This handler activates projection of constant force.
*******************************************************************************/
{
	memcpy(deviceState.userForce, pUserData, sizeof(hduVector3Dd));
	deviceState.isUserForceActive=1;

    return HD_CALLBACK_DONE;
}

HDCallbackCode HDCALLBACK resetHapticUserForce(void *pUserData)
/*******************************************************************************
 This handler deactivates projection of constant force.
*******************************************************************************/
{
	deviceState.isUserForceActive=0;

    return HD_CALLBACK_DONE;
}

HDCallbackCode HDCALLBACK setHapticAnchor(void *pUserData)
/*******************************************************************************
 This handler activates a spring force towards a given anchor.
*******************************************************************************/
{
	memcpy(deviceState.anchor, pUserData, sizeof(hduVector3Dd));
	deviceState.isAnchorActive=1;

    return HD_CALLBACK_DONE;
}

HDCallbackCode HDCALLBACK resetHapticAnchor(void *pUserData)
/*******************************************************************************
 This handler dactivates a spring force towards an anchor.
*******************************************************************************/
{
	deviceState.isAnchorActive=0;

    return HD_CALLBACK_DONE;
}

HDCallbackCode HDCALLBACK updateDeviceCallback(void *data)
/*******************************************************************************
 This handler updates deviceState and recomputes projected forces.
*******************************************************************************/
{
	/* Poczatek ramki haptycznej */
	hdBeginFrame(ghHD);
	int buttons=0;

	hdGetIntegerv(HD_CURRENT_BUTTONS, &buttons);

	deviceState.button1=(buttons & HD_DEVICE_BUTTON_1) ? 1 : 0;
	deviceState.button2=(buttons & HD_DEVICE_BUTTON_2) ? 1 : 0;

    hdGetDoublev(HD_CURRENT_POSITION, deviceState.position);
    hdGetDoublev(HD_CURRENT_TRANSFORM, deviceState.transform);
    hdGetDoublev(HD_CURRENT_FORCE, deviceState.force);

	hduVector3Dd force;
	hduVector3Dd uForce;
	hduVector3Dd aForce;
	memset(force, 0, sizeof(hduVector3Dd));
	memset(uForce, 0, sizeof(hduVector3Dd));
	memset(aForce, 0, sizeof(hduVector3Dd));

	if(deviceState.isUserForceActive) {
		memcpy(uForce, deviceState.userForce, sizeof(hduVector3Dd));

		double modForce=hduVecMagnitude(uForce);

		if(modForce>maxForce) {
			hduVecScaleInPlace(uForce, maxForce/modForce); 
		}
	}

	if(deviceState.isAnchorActive) {
		hduVector3Dd positionTanchor;

		hduVecSubtract(positionTanchor, deviceState.anchor, deviceState.position);

		double scale=k;

        hduVecScale(aForce, positionTanchor, scale);
	}

	hduVecAdd(deviceState.force, uForce, aForce);
	hdSetDoublev(HD_CURRENT_FORCE, deviceState.force);

	/* Koniec ramki. */
	hdEndFrame(ghHD);

	/* Sprawdzenie, czy nie wystapil blad. */
	HDErrorInfo error;
	if (HD_DEVICE_ERROR(error = hdGetError())) {
		hduPrintError(stderr, &error, "Error\n");

		if (hduIsSchedulerError(&error)) {
			return HD_CALLBACK_DONE;
		}
	}

	/* Ten callback ma byc wywolany rowniez w nastepnej iteracji schedulera. */
	return HD_CALLBACK_CONTINUE;
}

void initHD()
{
    HDErrorInfo error;
    ghHD = hdInitDevice(HD_DEFAULT_DEVICE);
    if (HD_DEVICE_ERROR(error = hdGetError())) 
    {
        hduPrintError(stderr, &error, "Failed to initialize haptic device");
        fprintf(stderr, "\nPress any key to quit.\n");
        getchar();
        exit(-1);
    }

    hdEnable(HD_FORCE_OUTPUT);

    hdStartScheduler();
    if (HD_DEVICE_ERROR(error = hdGetError()))
    {
        hduPrintError(stderr, &error, "Failed to start the scheduler");
        exit(-1);
    }
	calibrate_haptic();

	printf("Press any key\n");
	getchar();

    hUpdateDeviceCallback = hdScheduleAsynchronous(updateDeviceCallback, 0, HD_MAX_SCHEDULER_PRIORITY);
}

/*******************************************************************************
  Draws a 3D cursor for the haptic device using the current local transform,
  the workspace to world transform and the screen coordinate scale.
 *******************************************************************************/
void drawCursor(const HapticDisplayState *pState)
{
	static const double kCursorRadius = 0.5;
	static const double kCursorHeight = 1.5;
	static const int kCursorTess = 15;

	static GLuint gCursorDisplayList = 0;

	GLUquadricObj *qobj = 0;

	glPushAttrib(GL_CURRENT_BIT | GL_ENABLE_BIT | GL_LIGHTING_BIT);
	glPushMatrix();

	if (!gCursorDisplayList)
	{
		gCursorDisplayList = glGenLists(1);
		glNewList(gCursorDisplayList, GL_COMPILE);
		qobj = gluNewQuadric();

		gluCylinder(qobj, 0.0, kCursorRadius, kCursorHeight,
				kCursorTess, kCursorTess);
		glTranslated(0.0, 0.0, kCursorHeight);
		gluCylinder(qobj, kCursorRadius, 0.0, kCursorHeight / 5.0,
				kCursorTess, kCursorTess);

		gluDeleteQuadric(qobj);
		glEndList();
	}

	/* Apply the workspace view transform. */
	glMultMatrixd(workspacemodel);

	/* Apply the local transform of the haptic device. */
	glMultMatrixd(pState->transform);

	/* Apply the local cursor scale factor. */
	glScaled(gCursorScale, gCursorScale, gCursorScale);

	glEnable(GL_COLOR_MATERIAL);
	glColor3f(0.0, 0.5, 1.0);

	glCallList(gCursorDisplayList);

	glPopMatrix(); 
	glPopAttrib();
}

void updateWorkspace()
/*******************************************************************************
  Uses the current OpenGL viewing transforms to initialize a transform for the
  haptic device workspace so that it's properly mapped to world coordinates.
 *******************************************************************************/
{
	HDdouble screenTworkspace;
	GLdouble modelview[16];
	GLdouble projection[16];
    GLint viewport[4];

    glGetDoublev(GL_MODELVIEW_MATRIX, modelview);
    glGetDoublev(GL_PROJECTION_MATRIX, projection);
    glGetIntegerv(GL_VIEWPORT, viewport);

    /* Compute the transform for going from device coordinates to world
       coordinates based on the current viewing transforms. */
    hduMapWorkspaceModel(modelview, projection, workspacemodel);

    /* Compute the scale needed to display the cursor of a particular size
       in screen coordinates. */
    screenTworkspace = hduScreenToWorkspaceScale(
        modelview, projection, viewport, workspacemodel);

    gCursorScale = kCursorScreenSize * screenTworkspace;
}

void getModelCoord(t_coord *res, hduVector3Dd vec)
/*******************************************************************************
 Converts model coordinates to viewport (and haptic).
 *******************************************************************************/
{
	t_coord pos={vec[0], vec[1], vec[2], 0};

	res->x=pos.x*workspacemodel[0] + pos.y*workspacemodel[4] + pos.z*workspacemodel[8]  + workspacemodel[12];
	res->y=pos.x*workspacemodel[1] + pos.y*workspacemodel[5] + pos.z*workspacemodel[9]  + workspacemodel[13];
	res->z=pos.x*workspacemodel[2] + pos.y*workspacemodel[6] + pos.z*workspacemodel[10] + workspacemodel[14];
}

void getHapticCoord(hduVector3Dd res, t_coord vec)
/*******************************************************************************
 Converts viewport coordinates to model.
 *******************************************************************************/
{
	t_coord pos=vec;

	pos.x-=workspacemodel[12];
	pos.y-=workspacemodel[13];
	pos.z-=workspacemodel[14];

	double det=0;

	det+=workspacemodel[0] * workspacemodel[5] * workspacemodel[10];
	det+=workspacemodel[1] * workspacemodel[6] * workspacemodel[8];
	det+=workspacemodel[2] * workspacemodel[4] * workspacemodel[9];

	det-=workspacemodel[0] * workspacemodel[6] * workspacemodel[9];
	det-=workspacemodel[1] * workspacemodel[4] * workspacemodel[10];
	det-=workspacemodel[2] * workspacemodel[5] * workspacemodel[8];


	res[0]=pos.x*workspacemodel[0] + pos.y*workspacemodel[1] + pos.z*workspacemodel[2];
	res[1]=pos.x*workspacemodel[4] + pos.y*workspacemodel[5] + pos.z*workspacemodel[6];
	res[2]=pos.x*workspacemodel[8] + pos.y*workspacemodel[9] + pos.z*workspacemodel[10];

	double scale=pow(det, -2.0/3.0);

	res[0]*=scale;
	res[1]*=scale;
	res[2]*=scale;
}


